#ifndef _GLOBAL
#include "Globals.hlsli"
#endif

struct VertexInput
{
    uint myIndex : SV_VertexID;
};

struct VertexToPixel
{
    float4 myPosition : SV_POSITION;
    float4 myVSPosition : VS_POSITION;
    float2 myUV : UV;
};

struct PixelOutput
{
    float4 myColor : SV_TARGET;
};

struct DepthOutput
{
    float myDepth : SV_Depth;
};

Texture2D depthTexture      : register(t0);
Texture2D colorTexture      : register(t1);
Texture2D surfTexture       : register(t2);
Texture2D bumpTexture       : register(t3);
Texture2D emitTexture       : register(t4);

float LinearDepthCalc(float aDepthTextureSampleVal)
{
    return projectionB / (aDepthTextureSampleVal - projectionA);
}

float ProjectedDepth(VertexToPixel input)
{
    return depthTexture.Sample(pointClampSampler, input.myUV.xy).r;
}

float LinearDepth(VertexToPixel input)
{
    float zw = ProjectedDepth(input);
    return LinearDepthCalc(zw);
}

bool IsNotEmpty(VertexToPixel input)
{
    return ProjectedDepth(input) > 0;
}

float4 ViewPos(VertexToPixel input)
{
    float zw = ProjectedDepth(input);
    float3 viewRay = float3(input.myVSPosition.x, input.myVSPosition.y, 1.0f);
    float linearDepth = LinearDepthCalc(zw);
    return float4(viewRay * linearDepth, 1.0f * IsNotEmpty(input));
}

float4 WorldPosCalc(float4 aViewPos)
{
    float4 worldPosXY = mul(float4(aViewPos.xyz, 1.0f), fromCamera);
    return float4(worldPosXY.xyz, aViewPos.w);
}

float4 WorldPos(VertexToPixel input)
{
    float4 viewPos = ViewPos(input);
    return WorldPosCalc(viewPos);
}

float3 Albedo(VertexToPixel input)
{
    return colorTexture.Sample(pointClampSampler, input.myUV.xy).rgb;
}

float Alpha(VertexToPixel input)
{
    return colorTexture.Sample(pointClampSampler, input.myUV.xy).a;
}

float3 Normal(VertexToPixel input)
{
    return bumpTexture.Sample(pointClampSampler, input.myUV.xy).rgb;
}

float Height(VertexToPixel input)
{
    return bumpTexture.Sample(pointClampSampler, input.myUV.xy).a;
}

float AmbientOcclusion(VertexToPixel input)
{
    return surfTexture.Sample(pointClampSampler, input.myUV.xy).r;
}

float Roughness(VertexToPixel input)
{
    return surfTexture.Sample(pointClampSampler, input.myUV.xy).g;
}

float Metalness(VertexToPixel input)
{
    return surfTexture.Sample(pointClampSampler, input.myUV.xy).b;
}

float Mask(VertexToPixel input)
{
    return surfTexture.Sample(pointClampSampler, input.myUV.xy).a;
}

float3 Emissive(VertexToPixel input)
{
    return emitTexture.Sample(pointClampSampler, input.myUV.xy).rgb;
}

float3 Scattering(VertexToPixel input)
{
    return emitTexture.Sample(pointClampSampler, input.myUV.xy).a;
}