#include "..\Globals.hlsli"
#include "..\GBufferInput.hlsli"

cbuffer LightBuffer : register(b2)
{
    float4 lightPos;
    float3 spotLightColor;
    float lightRange;
    float3 lightDirection;
    float lightCone;
}

float3 Substance(VertexToPixel input)
{
    float3 albedo = Albedo(input).rgb;
    float metalness = Metalness(input);
    float3 substance = float3(0.04f, 0.04f, 0.04f) - (float3(0.04f, 0.04f, 0.04f) * metalness) + albedo * metalness;
    return substance;
}

float3 MetalnessAlbedo(VertexToPixel input)
{
    float3 albedo = Albedo(input);
    float metalness = Metalness(input);
    float3 metalnessAlbedo = albedo - (albedo * metalness);
    return metalnessAlbedo;
}

float3 Fresnel(VertexToPixel input, float3 aLightDirection)
{
    float3 toEye = normalize(cameraPos.xyz - WorldPos(input).xyz);
    float3 halfVector = normalize(aLightDirection + toEye);

    float3 substance = Substance(input).rgb;

    float LdotH = dot(aLightDirection, halfVector);
    LdotH = 1.0f - saturate(LdotH);
    LdotH = pow(LdotH, 5);
    float3 fresnel = LdotH * (1.0f - substance) + substance;

    return fresnel;
}

float Distribution(VertexToPixel input, float3 aLightDirection)
{
    float roughness = Roughness(input);
    float3 normal = Normal(input);

    float3 toEye = normalize(cameraPos.xyz - WorldPos(input).xyz);
    float3 halfVector = normalize(aLightDirection + toEye.xyz);

    float HdotN = saturate(dot(halfVector, normal));
    float m = roughness * roughness;
    float mm = m * m;
    float denominator = HdotN * HdotN * (mm - 1.0f) + 1.0f;
    float distribution = mm / (3.141593f * denominator * denominator);

    return distribution;
}

float Visibility(VertexToPixel input, float3 aLightDirection)
{
    float3 toEye = normalize(cameraPos.xyz - WorldPos(input).xyz);
    float roughness = Roughness(input);
    float roughnessRemapped = (roughness + 1.0f) / 2.0f;
    float3 normal = Normal(input);

    float NdotL = saturate(dot(normal, aLightDirection));
    float NdotV = saturate(dot(normal, toEye));

    float k = roughnessRemapped * roughnessRemapped * sqrt(3.1415f);
    float G1V = NdotV * (1.0f - k) + k;
    float G1L = NdotL * (1.0f - k) + k;
    float visibility = (NdotV * NdotL) / (G1V * G1L);

    return visibility;
}

PixelOutput main(VertexToPixel input)
{
	float3 lightColor = spotLightColor.rgb;
    float3 worldPosition = WorldPos(input).xyz;
    float3 toLight = lightPos.xyz - worldPosition.xyz;
    float toLightDistance = length(toLight);
    toLight = normalize(toLight);

    float3 metalnessAlbedo = MetalnessAlbedo(input).rgb;

    float3 normal = Normal(input);
    float lambertAttenuation = saturate(dot(normal, toLight));
    float linearAttenuation = saturate(1.0f - toLightDistance / lightRange);
    //float physicalAttenuation = 1.0f / (toLightDistance * toLightDistance);
    float coneAttenuation = 0;
    float spotLightFactor = abs(dot(toLight, lightDirection));
    float cutoff = cos(lightCone);
    if (spotLightFactor > cutoff)
    {
        coneAttenuation = (1.0 - (1.0 - spotLightFactor) * 1.0 / (1.0 - cutoff));
    }
    //float coneAttenuation = saturate((dot(toLight, lightDirection) - 6) / (5 - 6));
    float attenuation = lambertAttenuation * linearAttenuation * coneAttenuation;

    float3 fresnel = Fresnel(input, toLight).rgb;
    float distribution = Distribution(input, toLight);
    float visibility = Visibility(input, toLight);

    PixelOutput output;
    float3 diffuse = metalnessAlbedo * lightColor * attenuation * (float3(1.0f, 1.0f, 1.0f) - fresnel);
    float3 specularity = lightColor * attenuation * fresnel * distribution * visibility;
    output.myColor.rgb = diffuse + specularity;
    output.myColor.a = 1.0f * IsNotEmpty(input);

    return output;
}