#include "..\Globals.hlsli"
#include "..\ModelShaderStructs.hlsli"

TextureCube cubeMapTexture : register(t4);

cbuffer LightBuffer : register(b2)
{
    struct PointLightData
    {
        float4 myPosition;
        float3 myColor;
        float myRange;
    } myPointLight;

    float4 myToDirectionalLight;
    float4 myDirectionalLightColor;
    uint myCubeMapMipCount;
}

Texture2D colorTex : register(t0); // RGB = albedo,    A = Alpha
Texture2D surfTex : register(t1); // R   = AO,        G = roughness, B = metallic, A = blending mask
Texture2D bumpTex : register(t2); // RGB = normal,    A = height
Texture2D emitTex : register(t3); // RBG = emissive,  A = scattering

TextureCube cubeMap : register(t11);

PixelOutput Normal(VertexToPixel input)
{
    float3 texNormal = bumpTex.Sample(defaultClampSampler, input.myUV.xy).rgb;
    texNormal = (texNormal * 2) - 1;

    float3x3 tangentSpaceMatrix = float3x3(normalize(input.myBinormal.xyz), normalize(input.myTangent.xyz), normalize(input.myNormal.xyz));
    tangentSpaceMatrix = transpose(tangentSpaceMatrix);
    float3 pixelNormal = mul(tangentSpaceMatrix, texNormal.xyz);
    PixelOutput output;
    output.myColor = float4(pixelNormal, 0.0f);
    return output;
}

PixelOutput Substance(VertexToPixel input)
{
    float3 albedo = colorTex.Sample(defaultClampSampler, input.myUV).rgb;
    float metalness = surfTex.Sample(defaultClampSampler, input.myUV).g;
    float3 substance = float3(0.04f, 0.04f, 0.04f) - (float3(0.04f, 0.04f, 0.04f) * metalness) + albedo * metalness;

    PixelOutput output;
    output.myColor = float4(substance, 1.0f);
    return output;
}

PixelOutput Fresnel(VertexToPixel input)
{
    float3 toLight = myToDirectionalLight.xyz;
    float3 toEye = normalize(cameraPos.xyz - input.myWorldPosition.xyz);
    float3 halfVector = normalize(toLight + toEye);

    float3 substance = Substance(input).myColor.rgb;

    float LdotH = dot(toLight, halfVector);
    LdotH = 1.0f - saturate(LdotH);
    LdotH = pow(LdotH, 5);
    float3 fresnel = LdotH * (1.0f - substance) + substance;

    PixelOutput output;
    output.myColor = float4(fresnel.rgb, 1.0f);
    return output;
}

PixelOutput MetalnessAlbedo(VertexToPixel input)
{
    float3 albedo = colorTex.Sample(defaultClampSampler, input.myUV).rgb;
    float metalness = surfTex.Sample(defaultClampSampler, input.myUV).b;
    float3 metalnessAlbedo = albedo - (albedo * metalness);
    PixelOutput output;
    output.myColor = float4(metalnessAlbedo, 1.0f);
    return output;
}

PixelOutput Lambert(VertexToPixel input)
{
    float3 toLight = myToDirectionalLight.xyz;
    float3 normal = Normal(input).myColor.rgb;
    
    float3 lambert = saturate(dot(toLight, normal));

    PixelOutput output;
    output.myColor = float4(lambert.rrr, 1.0f);
    return output;
}

PixelOutput ReflectionFresnel(VertexToPixel input)
{
    float roughness = surfTex.Sample(defaultClampSampler, input.myUV).g;
    float3 substance = Substance(input).myColor.rgb;
    float3 normal = Normal(input).myColor.rgb;
    float3 toEye = normalize(cameraPos.xyz - input.myWorldPosition.xyz);

    float VdotN = dot(toEye, normal);
    VdotN = 1.0f - saturate(VdotN);
    VdotN = pow(VdotN, 5);

    float3 fresnel = VdotN * (1.0f - substance);
    fresnel = fresnel / (6 - 5 * roughness) + substance;

    PixelOutput output;
    output.myColor = float4(fresnel, 1.0f);
    return output;
}

PixelOutput AmbientDiffuse(VertexToPixel input)
{
    float3 normal = Normal(input).myColor.rgb;
    float3 metalnessAlbedo = MetalnessAlbedo(input).myColor.rgb;
    float ambientOcclusion = surfTex.Sample(defaultClampSampler, input.myUV).r;

    //float3 environmentLight = float3(0.1f, 0.1f, 0.1f); 
    float3 environmentLight = cubeMapTexture.SampleLevel(defaultClampSampler, normal, myCubeMapMipCount - 2).rgb;
    float3 fresnel = ReflectionFresnel(input).myColor.rgb;

    PixelOutput output;
    output.myColor.rgb = metalnessAlbedo * environmentLight * ambientOcclusion * (float3(1.0f, 1.0f, 1.0f) - fresnel);
    output.myColor.a = 1.0f;
    return output;
}

float RoughToSPow(float fRoughness)
{
    return (2.0f / (fRoughness * fRoughness)) - 2.0f;
}
static const float k0 = 0.00098f;
static const float k1 = 0.9921f;
static const float fakeLysMaxSpecularPower = (2.0f / (0.0014f * 0.0014f)) - 2.0f;
static const float fMaxT = (exp2(-10.0f / sqrt(fakeLysMaxSpecularPower)) - k0) / k1;
float GetSpecPowToMip(float fSpecPow, int nMips)
{
    float fSmulMaxT = (exp2(-10.0f / sqrt(fSpecPow)) - k0) / k1;
    return float(nMips - 1 - 0) * (1.0f - clamp(fSmulMaxT / fMaxT, 0.0f, 1.0f));
}

PixelOutput AmbientSpecular(VertexToPixel input)
{
    float3 normal = Normal(input).myColor.rgb;
    float roughness = surfTex.Sample(defaultClampSampler, input.myUV).g;
    float ambientOcclusion = surfTex.Sample(defaultClampSampler, input.myUV).r;

    float3 toEye = normalize(cameraPos.xyz - input.myWorldPosition.xyz);
    float3 reflectionVector = -reflect(toEye, normal);

    float fakeLysSpecularPower = RoughToSPow(roughness);
    float lysMipMap = GetSpecPowToMip(fakeLysSpecularPower, 10);

    //float3 environmentLight = float3(0.1f, 0.1f, 0.1f); 
    float3 environmentLight = cubeMapTexture.SampleLevel(defaultClampSampler, normal, myCubeMapMipCount - 2).rgb;
    float3 fresnel = ReflectionFresnel(input).myColor.rgb;

    PixelOutput output;
    output.myColor.rgb = environmentLight * ambientOcclusion * fresnel;
    output.myColor.a = 1.0f;
    return output;
}

PixelOutput DirectDiffuse(VertexToPixel input)
{
    float3 lightColor = myDirectionalLightColor.rgb;

    float3 metalnessAlbedo = MetalnessAlbedo(input).myColor.rgb;

    float3 lambert = Lambert(input).myColor.rgb;
    float3 fresnel = Fresnel(input).myColor.rgb;

    PixelOutput output;
    output.myColor.rgb = metalnessAlbedo * lightColor * lambert * (float3(1.0f, 1.0f, 1.0f) - fresnel);
    output.myColor.a = 1.0f;
    return output;
}

PixelOutput Distribution(VertexToPixel input)
{
    float roughness = surfTex.Sample(defaultClampSampler, input.myUV).g;
    float3 normal = Normal(input).myColor.rgb;

    float3 toEye = normalize(cameraPos.xyz - input.myWorldPosition.xyz);
    float3 halfVector = normalize(myToDirectionalLight.xyz + toEye.xyz);

    float HdotN = saturate(dot(halfVector, normal));
    float m = roughness * roughness;
    float mm = m * m;
    float denominator = HdotN * HdotN * (mm - 1.0f) + 1.0f;
    float distribution = mm / (3.141593f * denominator * denominator);

    PixelOutput output;
    output.myColor.rgb = distribution;
    output.myColor.a = 1.0f;
    return output;
}

PixelOutput Visibility(VertexToPixel input)
{
    float3 toLight = myToDirectionalLight.xyz;
    float3 toEye = normalize(cameraPos.xyz - input.myWorldPosition.xyz);
    float roughness = surfTex.Sample(defaultClampSampler, input.myUV).g;
    float roughnessRemapped = (roughness + 1.0f) / 2.0f;
    float3 normal = Normal(input).myColor.rgb;

    float NdotL = saturate(dot(normal, toLight));
    float NdotV = saturate(dot(normal, toEye));

    float k = roughnessRemapped * roughnessRemapped * sqrt(3.1415f);
    float G1V = NdotV * (1.0f - k) + k;
    float G1L = NdotL * (1.0f - k) + k;
    float visibility = (NdotV * NdotL) / (G1V * G1L);

    PixelOutput output;
    output.myColor.rgb = visibility.rrr;
    output.myColor.a = 1.0f;
    return output;
}

PixelOutput DirectSpecular(VertexToPixel input)
{
    float3 lightColor = myToDirectionalLight.rgb;
    float3 lambert = Lambert(input).myColor.rgb;
    float3 fresnel = Fresnel(input).myColor.rgb;
    float3 distribution = Distribution(input).myColor.rgb;
    float3 visibility = Visibility(input).myColor.rgb;

    PixelOutput output;
    output.myColor.rgb = lightColor * lambert * fresnel * distribution * visibility;
    output.myColor.a = 1.0f;
    return output;
}

PixelOutput main(VertexToPixel input)
{
    float3 ambientDiffuse = AmbientDiffuse(input).myColor.rgb;
    float3 ambientSpecular = AmbientSpecular(input).myColor.rgb;
    float3 directDiffuse = DirectDiffuse(input).myColor.rgb;
    float3 directSpecular = DirectSpecular(input).myColor.rgb;
    float3 emissive = emitTex.Sample(defaultClampSampler, input.myUV).rgb;

    PixelOutput output;
    output.myColor.rgb = ambientDiffuse + ambientSpecular + directDiffuse + directSpecular + emissive;
    output.myColor.a = 1.0f;
    return output;
}